#!/usr/bin/perl
#------------------------------------------------------------------------------
#    SCB2 - Simple Comment Book II
#    Copyright (c) 1999-2003 Arne Bochem <ascb@arneb.de>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#------------------------------------------------------------------------------

use strict;
use CGI qw(:standard);
use CGI::Carp qw(fatalsToBrowser);
use POSIX qw(strftime);
eval { use Fcntl qw(:flock); };

#------------------------------------------------------------------------------
# Start of configuration
my %cfg;

# Email program
$cfg{'mail_program'} = '/bin/mail';

# Email address
$cfg{'email'} = 'someuser@somewhere.tld';

# Send entries by email? (1|0)
$cfg{'send_entries'} = 0;

# Display IP address? (1|0)
$cfg{'display_ip'} = 1;

# Timestamp format
$cfg{'timestamp'} = '%Y-%m-%d %X';

# Name of data file
$cfg{'data_file'} = 'scb2.data';

# HTML header
$cfg{'html_header'} = <<EOF;
<?xml version="1.0" encoding="iso-8859-1"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN" "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">

<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en">
  <head>
    <meta name="robots" content="noindex, follow" />
    <meta name="generator" content="Simple Comment Book II" />
    <title>Simple Comment Book II</title>
  </head>
  <body>
    <h1>Simple Comment Book II</h1>
EOF

# HTML for success message
$cfg{'html_success'} = <<EOF;
    <h2>Success</h2>
    <p>Entry successfully saved:</p>
EOF

# HTML for failure message
$cfg{'html_failure'} = <<EOF;
    <h2>Failure</h2>
    <p>Entry not saved. Invalid form data input.</p>
EOF

# HTML for form
$cfg{'html_form'} = <<EOF;
    <h2>Create Entry</h2>
    <form method="post" action="scb2.pl">
      <table>
        <tr>
          <td style="text-align: right;">
            Name:
          </td>
          <td>
            <input type="text" name="fld_nm" />
          </td>
        </tr>
        <tr>
          <td style="text-align: right;">
            Email address:
          </td>
          <td>
            <input type="text" name="fld_ml" />
          </td>
        </tr>
        <tr>
          <td style="text-align: right;">
            Homepage:
          </td>
          <td>
            <input type="text" name="fld_rl" />
          </td>
        </tr>
        <tr>
          <td style="text-align: right;">
            Text:
          </td>
          <td>
            <textarea name="fld_ntry" rows="5" cols="50"></textarea>
          </td>
        </tr>
        <tr>
          <td style="text-align: right;">
            <input type="submit" value="Write" />
          </td>
          <td>
            <input type="reset" value="Clear" />
          </td>
        </tr>
      </table>
    </form>
    <h2>Entries</h2>
EOF

# HTML template for entry
$cfg{'html_entry'} = <<EOF;
    <div style="width: 100%; border: 1px black solid; padding: 0px; margin-bottom: 10px;">
      <p style="padding: 3px; margin: 0px; width: auto; background-color: silver; color: black; border-bottom: 1px gray solid;">
        <!--IDENTITY-->
      </p>
      <p style="padding: 3px; width: auto; margin: 0px; text-align: justify;">
        <!--BODY-->
      </p>
    </div>
EOF

# HTML footer
$cfg{'html_footer'} = <<EOF;
  </body>
</html>
EOF

# Running on a windows *cough* server
# $cfg{'windows'} = 1;

# End of configuration
#------------------------------------------------------------------------------
# Real code starts here

# Output header
print "Content-type: text/html\n\n$cfg{'html_header'}";

# Do what shall be done
param('fld_ntry') ? write_new_entry() : display_entries();

# Output footer
print_footer();

# Remove useless whitespace.
sub trim ($)
{
  my $text = shift;
  $text =~ s/^\s*//g;
  $text =~ s/\s*$//g;
  return($text);
}

# Check if email address has valid format
sub is_email ($)
{
  my $email = shift;
	return(1) if ($email =~ /^[\w\.\-_]+?\@[\w\.\-]+?\.\w{2,}$/);
	return(0);
}

# Check if url has valid format
sub is_http_url ($)
{
  my $url = shift;
  my $charclass = '[a-zA-Z0-9\\.\\-]{2,}';
  my $charclass2 = '[a-zA-Z0-9\\.\\-\\%\\_\\?\\&\\/\\+\\=\\~\\;]*';
  return(1) if ($url =~ /^https?:\/\/$charclass\.$charclass(?:\/$charclass2)?$/);
  return(0);
}

# Check if something has substance - could still be inane drivel though
sub has_substance ($)
{
	my $text = shift;
	return(0) if ($text =~ /^[\s\-\_\.]*$/s);
	return(1);
}

# Take care of stray HTML tags and htmlize newlines
sub htmlize ($)
{
	my $text = shift;
	$text =~ s/</&lt;/g;
	$text =~ s/>/&gt;/g;
	$text =~ s#\r\n|\n|\r#<br />#g;
	return($text);
}

# Remove newlines and similar stuff
sub remove_newlines ($)
{
	my $text = shift;
	$text =~ s/[\r\n]//g;
	return($text);
}

# Try to confound spam spiders
sub obfuscate_email ($)
{
	my $email = shift;
	$email =~ s/\@/ at /g;
	$email =~ s/\./ dot /g;
	return($email);
}

# Append entry to data file
sub append_entry ($$$)
{
	my ($data_file, $body, $identity) = @_;
	
	# Open and lock the data file
	open(DATAFILE, '>>'.$data_file) || die("Could not open data file ($data_file): $!");
	(flock(DATAFILE, LOCK_EX) || die("Could not lock data file ($data_file): $!")) unless ($cfg{'windows'});

	# Seek to the end - just in case
	seek(DATAFILE, 0, 2);

	# Print newline, body, newline, identity
	print DATAFILE "\n$body\n$identity";

	# Close the file, everything ought to be fine
	close(DATAFILE) || die("Could not close data file ($data_file): $!");
}

# Mail entry to owner
sub mail_to_owner ($$)
{
	my ($email, $data) = @_;

	# Check validity of owner email address
	die("This should not happen. Owner email address invalid: $email") unless (is_email($email));

	# Open pipe to mail program
	open(MAILPIPE, '|'.$cfg{'mail_program'}.' '.$email) || die("Could not open mail program ($cfg{'mail_program'}): $!");

	# Write mail header
	print MAILPIPE
		"To: $email\n",
		"From: Simple Comment Book II <$email>\n",
		"Subject: New entry\n",
		"MIME-Version: 1.0\n",
		"Content-Type: text/plain; charset=iso-8859-1\n",
		"Content-Transfer-Encoding: 8bit\n\n";
	
	# Write mail body
	foreach my $field_name (keys %{$data})
	{
		print MAILPIPE
			"$field_name:\n",
			$data->{$field_name}."\n",
			("-"x40)."\n";
	}

	# Done, close pipe
	close(MAILPIPE) || die("Could not close mail program ($cfg{'mail_program'}): $!");
}

# Apply entry HTML template to data
sub apply_template ($$$)
{
	my ($template, $identity, $body) = @_;

	# Replace placeholders with content
	$template =~ s/<!--IDENTITY-->/$identity/gi;
	$template =~ s/<!--BODY-->/$body/gi;

	return($template);
}

# Slurp content of data file
sub slurp_data_file ($)
{
	my $data_file = shift;

	# Open data file
	open(DATAFILE, '<'.$data_file) || die("Could not open data file ($data_file): $!");
	
	# Slurp content into array
	my @content = <DATAFILE>;

	# Done - close
	close(DATAFILE);

	return(@content);
}

# Try to save new entry
sub write_new_entry
{
	my $name  = remove_newlines(trim(param('fld_nm')));
	my $email = trim(param('fld_ml'));
	my $url   = trim(param('fld_rl'));
	my $body  = trim(param('fld_ntry'));
	my $time  = strftime($cfg{'timestamp'}, localtime(time()));
	my $ip    = $ENV{'REMOTE_ADDR'};

	# We do not want any entries without substance
	unless (has_substance($name) and has_substance($body))
	{
		print $cfg{'html_failure'};
		return;
	}
	# And we do not want invalid information either
	$email = '' unless (is_email($email));
	$url = '' unless (is_http_url($url));

	# Let us create the identity line
	my $identity;

	$identity .= '<a href="'.$url.'">' if ($url);
	$identity .= htmlize($name);
	$identity .= '</a>' if ($url);
	$identity .= ' ('.$ip.')' if ($cfg{'display_ip'});
	$identity .= ' &lt;'.obfuscate_email($email).'&gt;' if($email);
	$identity .= ' - '.$time;

	# Append identity and entry body to data file
	append_entry($cfg{'data_file'}, htmlize($body), $identity);

	# Mail entry to owner, if enabled
	mail_to_owner($cfg{'email'}, {'Name' => $name, 'Email' => $email, 'Url' => $url, 'Text' => $body, 'Timestamp' => $time, 'IP' => $ip}) if ($cfg{'send_entries'});

	# Output success message and display entry
	print $cfg{'html_success'}.apply_template($cfg{'html_entry'}, $identity, htmlize($body));
}

# Insert copyright notice into footer and print it
sub print_footer
{
	my $copyright_notice = "\t\t".'<p style="color: silver; border: 1px black solid; background-color: #444444; font-family: Verdana, Helvetica, sans-serif; font-size: smaller; width: 100%; padding: 1px; text-align: center;"><a href="http://www.arneb.de/" style="color: white; background-color: #444444; text-decoration: underline;">Simple Comment Book II</a> is &copy;2003 by Arne Bochem &lt;scb2 at arneb dot de&gt;</p>'."\n";
	$cfg{'html_footer'} =~ s#(\n*)(\s*</body>)#$1.$copyright_notice.$2#gie;
	print $cfg{'html_footer'};
}

# Display entries
sub display_entries
{
	print $cfg{'html_form'};

	# Slrup data file
	my @entry_data_unsorted = slurp_data_file($cfg{'data_file'});

	# Put elements in right order and remove empty ones
	my @entry_data;
	while (@entry_data_unsorted)
	{
		my $element = pop(@entry_data_unsorted);
		next unless ($element =~ /\S/);
		push(@entry_data, $element);
	}

	# Output entries
	while (scalar(@entry_data) > 1)
	{
		# Get entry data
		my $identity = shift(@entry_data);
		my $body = shift(@entry_data);

		# Print it
		print apply_template($cfg{'html_entry'}, $identity, $body);
	}
}
